r"""
Arithmetic on the Jacobian

This module implements the group operation in the Picard group of a
hyperelliptic curve, represented as divisors in Mumford
representation, using Cantor's algorithm.

AUTHORS:

- Sabrina Kunzweiler, Gareth Ma, Giacomo Pope (2024): adapt to smooth model
"""

# ****************************************************************************
#       Copyright (C) 2025 Sabrina Kunzweiler <sabrina.kunzweiler@math.u-bordeaux.fr>
#                     2025 Gareth Ma <grhkm21@gmail.com>
#                     2025 Giacomo Pope <giacomopope@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from __future__ import annotations

from typing import TYPE_CHECKING, Self

if TYPE_CHECKING:
    from collections.abc import Iterator

    from sage.rings.integer import Integer

from sage.groups.generic import order_from_multiple
from sage.misc.cachefunc import cached_method
from sage.rings.finite_rings.finite_field_base import FiniteField as FiniteField_generic
from sage.rings.polynomial.polynomial_element import Polynomial
from sage.schemes.generic.morphism import SchemeMorphism
from sage.structure.element import AdditiveGroupElement
from sage.structure.richcmp import richcmp


class MumfordDivisorClassField(AdditiveGroupElement, SchemeMorphism):
    r"""
    An element of a Jacobian defined over a field, i.e. in
    `J(K) = \mathrm{Pic}^0_K(C)`.
    """

    def __init__(self, parent, u, v, check=True) -> None:
        """
        Create an element of the Jacobian of a hyperelliptic curve.

        EXAMPLES::

            sage: R.<x> = GF(19)[]
            sage: H = HyperellipticCurve(x^5 + x, x^2 + 1)
            sage: J = H.jacobian()
            sage: D = J(x^2 + 14*x + 16, 3*x + 4); D # indirect doctest
            (x^2 + 14*x + 16, 3*x + 4)
        """
        SchemeMorphism.__init__(self, parent)

        if not isinstance(u, Polynomial) or not isinstance(v, Polynomial):
            raise TypeError(f"arguments u={u} and v={v} must be polynomials")

        # TODO:
        # 1. allow elements of the base field as input
        #   (in particular something like (u,v) = (x-alpha, 0))

        # Ensure the divisor is valid
        if check:
            f, h = parent.curve().hyperelliptic_polynomials()
            assert (v**2 + v * h - f) % u == 0, (
                f"u={u}, v={v} do not define a divisor on the Jacobian"
            )

            # TODO: should we automatically do reduction here if the degree of u is
            # too large?

        self._parent = parent
        self._u: Polynomial = u
        self._v: Polynomial = v

    def scheme(self):
        r"""
        Return the scheme this morphism maps to; or, where this divisor lives.

        .. WARNING::

           Although a pointset is defined over a specific field, the
           scheme returned may be over a different (usually smaller)
           field.  The example below demonstrates this: the pointset
           is determined over a number field of absolute degree 2 but
           the scheme returned is defined over the rationals.

        EXAMPLES::

            sage: x = QQ['x'].gen()
            sage: f = x^5 + x
            sage: H = HyperellipticCurve(f)
            sage: F.<a> = NumberField(x^2 - 2, 'a')                                     # needs sage.rings.number_field
            sage: J = H.jacobian()(F); J                                                # needs sage.rings.number_field
            Abelian group of points over Number Field in a with defining polynomial x^2 - 2 on Jacobian of Hyperelliptic Curve over Rational Field defined by y^2 = x^5 + x


        ::

            sage: P = J(H.lift_x(F(1)))                                                 # needs sage.rings.number_field
            sage: P.scheme()                                                            # needs sage.rings.number_field
            Jacobian of Hyperelliptic Curve over Rational Field defined by y^2 = x^5 + x
        """
        return self.codomain()

    def _repr_(self) -> str:
        r"""
        Return the Mumford presentation of the divisor class.

        EXAMPLES::

            sage: R.<x> = PolynomialRing(QQ)
            sage: H = HyperellipticCurve(x^5 + 2*x^3 + x, x^3 + 1)
            sage: P = H([0,-1])
            sage: Q = H([0,0])
            sage: J = H.jacobian()
            sage: D = J(P,Q); D # indirect doctest
            (x^2, -x - 1 : 0)
        """
        return f"({self._u}, {self._v})"

    def uv(self) -> tuple[Polynomial, Polynomial]:
        r"""
        Return the `u` and `v` component of this Mumford divisor.

        EXAMPLES::

            sage: x = polygen(GF(1993))
            sage: H = HyperellipticCurve(x^7 + 3 * x + 1)
            sage: J = H.jacobian(); J
            Jacobian of Hyperelliptic Curve over Finite Field of size 1993 defined by y^2 = x^7 + 3*x + 1
            sage: u, v = x^3 + 1570*x^2 + 1930*x + 81, 368*x^2 + 1478*x + 256
            sage: P = J(u, v); P
            (x^3 + 1570*x^2 + 1930*x + 81, 368*x^2 + 1478*x + 256)
            sage: P.uv() == (u, v)
            True
        """
        return (self._u, self._v)

    def _richcmp_(self, other, op) -> bool:
        r"""
        Method for rich comparison.

        TESTS::

            sage: x = polygen(GF(23))
            sage: H = HyperellipticCurve(x^7 + x + 1)
            sage: J = H.jacobian(); J
            Jacobian of Hyperelliptic Curve over Finite Field of size 23 defined by y^2 = x^7 + x + 1
            sage: P = J.random_element()
            sage: P == P
            True
            sage: P != P
            False

            sage: Z = J(0); Z
            (1, 0)
            sage: Z == 0
            True
            sage: Z != 0
            False
        """
        return richcmp(tuple(self), tuple(other), op)

    def __iter__(self) -> Iterator[Polynomial]:
        r"""
        TESTS:

        Indirect tests::

            sage: R.<x> = GF(13)[]
            sage: H = HyperellipticCurve(x^5 - x + 1)
            sage: J = H.jacobian()
            sage: P = J(x + 5, 12)
            sage: list(P)
            [x + 5, 12]
            sage: tuple(P)
            (x + 5, 12)
        """
        yield from [self._u, self._v]

    def __getitem__(self, n) -> Polynomial:
        r"""
        Return the n-th item in the Mumford presentation of the divisor.

        TESTS::

            sage: R.<x> = GF(23)[]
            sage: H = HyperellipticCurve(x^5 + x + 1, x^2 + x +  1)
            sage: J = H.jacobian()
            sage: D = J(x^2 + 21*x + 10, 14*x + 22)
            sage: D[0] # indirect doctest
            x^2 + 21*x + 10
            sage: D[1] # indirect doctest
            14*x + 22
        """
        return (self._u, self._v)[n]

    def __hash__(self) -> int:
        r"""
        Compute the hash value of this element.

        TESTS::

            sage: R.<x> = GF(23)[]
            sage: H = HyperellipticCurve(x^5 + x + 1, x^2 + x +  1)
            sage: J = H.jacobian()
            sage: D = J(x^2 + 21*x + 10, 14*x + 22)
            sage: hash(D) == hash(2*D)
            False
        """
        return hash(tuple(self))

    def __reduce__(self) -> tuple:
        r"""
        Return data for pickling.

        TESTS::

            sage: x = polygen(GF(5))
            sage: H = HyperellipticCurve(x^5 + 3*x + 1)
            sage: J = H.jacobian()
            sage: P = J.random_element()
            sage: loads(dumps(P)) == P
            True
        """
        return (self._parent, tuple(self))

    def __bool__(self) -> bool:
        r"""
        Return ``True`` if this is not the zero element of the Jacobian.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^2 + x, x^3 + 1)
            sage: J = H.jacobian()
            sage: bool(J(x+1,0))
            True
            sage: bool(J.zero())
            False

        ::

            sage: x = polygen(GF(419))
            sage: C = HyperellipticCurve(x^5 + x)
            sage: bool(C.jacobian().zero())
            False
        """
        return not (self._u.is_one() and self._v.is_zero())

    @cached_method
    def order(self) -> Integer:
        r"""
        Returns the order of self.
        This is only implemented over finite fields.

        EXAMPLES::

            sage: K = FiniteField(7)
            sage: R.<x> = K[]
            sage: H = HyperellipticCurve(x^6 + 3*x + 2)
            sage: JK = Jacobian(H)(K)
            sage: D = JK(x^2 + 5*x + 6, 6*x + 3)
            sage: D.order()
            38
        """
        if not isinstance(self.base_ring(), FiniteField_generic):
            raise NotImplementedError(
                "this is only implemented for Jacobians over a finite field"
            )
        n = self.parent().order()
        return order_from_multiple(self, n)

    def degree(self) -> Integer:
        r"""
        Returns the degree of the affine part of the divisor.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(6*x^5 + 9*x^4 - x^3 - 3*x^2, 1)
            sage: J = H.jacobian()
            sage: J.zero().degree()
            0
            sage: J(x, 0).degree()
            1
            sage: J(x^2 + 1/2*x, -x - 1).degree()
            2
        """
        return self._u.degree()

    def _add_(self, other) -> Self:
        r"""
        Add `other` to the divisor `self`.

        TESTS::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^5 - x^4 + x^2 - x, 1)
            sage: J = H.jacobian()
            sage: D1 = J(x^2 + x, x)
            sage: D2 = J(x^2, x - 1)
            sage: D1 + D2
            (x^2 + x, -1)

            sage: Jp = J.change_ring(GF(13))
            sage: D1, D2, D3 = [Jp.random_element() for i in range(3)]
            sage: D1 + D2 + D3 == D3 + D1 + D2
            True
        """
        # Collect data from HyperellipticCurve
        H = self.parent().curve()
        g = H.genus()

        # Extract out mumford coordinates
        u1, v1 = self.uv()
        u2, v2 = other.uv()

        # Step one: cantor composition of the two divisors
        u3, v3 = self._parent.cantor_composition(u1, v1, u2, v2)

        # Step two: cantor reduction of the above to ensure
        # the degree of u is smaller than g + 1
        while u3.degree() > g:
            u3, v3 = self._parent.cantor_reduction(u3, v3)
        v3 = v3 % u3

        return self._parent(u3, v3, check=False)

    def _neg_(self) -> Self:
        r"""
        Multiply the divisor by `[-1]`.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^5 + 2*x^4 + 3*x^3 + 2*x^2 + x, 1)
            sage: P = H(1,0,0)
            sage: Q = H(0,0)
            sage: J = H.jacobian()
            sage: D = J(P,Q); D
            (x, -1)
            sage: -D #indirect doctest
            (x, 0)
            sage: J(Q,P) == - D #indirect doctest
            True
        """
        _, h = self.parent().curve().hyperelliptic_polynomials()
        u0, v0 = self.uv()

        if h.is_zero():
            return self._parent(u0, -v0, check=False)

        v1 = (-h - v0) % u0
        return self._parent(u0, v1, check=False)


# =======================================================================
# Child classes for representation for ramified, inert and split models
# =======================================================================


class MumfordDivisorClassFieldRamified(MumfordDivisorClassField):
    def __init__(self, parent, u, v, check=True) -> None:
        r"""
        Create an element of the Jacobian of a ramified
        hyperelliptic curve.

        TESTS::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(x^7 + 3*x + 2, x^2 + 1)
            sage: H.is_ramified()
            True
            sage: J = Jacobian(H)
            sage: D = J(x^3 + x^2 + 3*x + 1, 4*x^2 + 2*x + 3)
            sage: type(D)
            <class 'sage.schemes.hyperelliptic_curves.jacobian_morphism.MumfordDivisorClassFieldRamified'>
        """
        if not parent.curve().is_ramified():
            raise TypeError("hyperelliptic curve must be ramified")

        super().__init__(parent, u, v, check=check)


class MumfordDivisorClassFieldInert(MumfordDivisorClassField):
    def __init__(self, parent, u, v, check=True) -> None:
        r"""
        Create an element of the Jacobian of an inert
        hyperelliptic curve.

        TESTS::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(2*x^6 + 1)
            sage: H.is_inert()
            True
            sage: J = Jacobian(H)
            sage: D = J(x^2 + x + 4, 4*x)
            sage: type(D)
            <class 'sage.schemes.hyperelliptic_curves.jacobian_morphism.MumfordDivisorClassFieldInert'>
        """
        if not parent.curve().is_inert():
            raise TypeError("hyperelliptic curve must be inert")

        if u.degree() % 2:
            raise ValueError(f"mumford coordinate {u} must have even degree")

        g = parent.curve().genus()
        self._n = (g - u.degree()) // 2
        super().__init__(parent, u, v, check=check)

    def __repr__(self) -> str:
        """
        Return a representation of the element.

        TESTS::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(2*x^6 + 1)
            sage: H.is_inert()
            True
            sage: J = Jacobian(H)
            sage: D = J(x^2 + x + 4, 4*x); D # indirect doctest
            (x^2 + x + 4, 4*x : 0)
        """
        return f"({self._u}, {self._v} : {self._n})"

    def _add_(self, *args, **kwds):
        r"""
        Thin wrapper around :meth:`MumfordDivisorClassField._add_` which
        fails in case of odd genus since that case is not implemented.

        EXAMPLES::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(2*x^6 + 1)
            sage: H.genus()
            2
            sage: H.is_inert()
            True
            sage: J = Jacobian(H)
            sage: D = J(x^2 + x + 4, 4*x)
            sage: D + D
            (x^2 + x + 4, x : 0)

        ::

            sage: H = HyperellipticCurve(2*x^8 + 1)
            sage: H.genus()
            3
            sage: H.is_inert()
            True
            sage: J = Jacobian(H)
            sage: D = J(x^2 + 3, 3*x)
            sage: D + D
            Traceback (most recent call last):
            ...
            NotImplementedError: unable to perform arithmetic for inert models of odd genus; consider extending the base field to adjoin the points at infinity
        """
        if self.parent().curve().genus() % 2:
            raise NotImplementedError('unable to perform arithmetic for inert models of odd genus; '
                                      'consider extending the base field to adjoin the points at infinity')
        return super()._add_(*args, **kwds)

    def _neg_(self, *args, **kwds):
        r"""
        Thin wrapper around :meth:`MumfordDivisorClassField._neg_` which
        fails in case of odd genus since that case is not implemented.

        EXAMPLES::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(2*x^6 + 1)
            sage: H.genus()
            2
            sage: H.is_inert()
            True
            sage: J = Jacobian(H)
            sage: D = J(x^2 + x + 4, 4*x)
            sage: -D
            (x^2 + x + 4, x : 0)

        ::

            sage: H = HyperellipticCurve(2*x^8 + 1)
            sage: H.genus()
            3
            sage: H.is_inert()
            True
            sage: J = Jacobian(H)
            sage: D = J(x^2 + 3, 3*x)
            sage: -D
            Traceback (most recent call last):
            ...
            NotImplementedError: unable to perform arithmetic for inert models of odd genus; consider extending the base field to adjoin the points at infinity
        """
        if self.parent().curve().genus() % 2:
            raise NotImplementedError('unable to perform arithmetic for inert models of odd genus; '
                                      'consider extending the base field to adjoin the points at infinity')
        return super()._neg_(*args, **kwds)


class MumfordDivisorClassFieldSplit(MumfordDivisorClassField):
    def __init__(self, parent, u, v, n=0, check=True) -> None:
        r"""
        Create an element of the Jacobian of a split
        hyperelliptic curve.

        TESTS::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(x^6 + 1)
            sage: H.is_split()
            True
            sage: J = Jacobian(H)
            sage: D = J(x^2 + 3, 3)
            sage: type(D)
            <class 'sage.schemes.hyperelliptic_curves.jacobian_morphism.MumfordDivisorClassFieldSplit'>
        """
        if not parent.curve().is_split():
            raise TypeError("hyperelliptic curve must be split")

        # Ensure the weight is set correctly
        g = parent.curve().genus()
        assert 0 <= n <= (g - u.degree())
        self._n = n
        self._m = g - u.degree() - n

        super().__init__(parent, u, v, check=check)

    def __repr__(self) -> str:
        r"""
        Return a representation of the element.

        TESTS::

            sage: R.<x> = GF(5)[]
            sage: H = HyperellipticCurve(x^6 + 1)
            sage: J = Jacobian(H)
            sage: D = J(x^2 + 3, 3); D
            (x^2 + 3, 3 : 0)
        """
        return f"({self._u}, {self._v} : {self._n})"

    def __bool__(self) -> bool:
        r"""
        Return ``True`` if this is not the zero element of the Jacobian.

        TESTS:

        We consider a genus-3 hyperelliptic curve with two points
        at infinity ``P_0``, ``P_1``. Elements of the Jacobian are
        represented as ``[D - 2P_0 - P_1]``.
        In particular the zero element has Mumford presentation
        `(1, 0 : 2)`::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^8  + 1)
            sage: J = Jacobian(H)
            sage: D = J(1,0,2); D
            (1, 0 : 2)
            sage: D.is_zero()
            True

        There are more elements that are only supported at infinity,
        but are non-zero::

            sage: [P0,P1] = H.points_at_infinity()
            sage: D = J(P0,P1); D
            (1, 0 : 3)
            sage: D.is_zero()
            False
        """
        g = self._parent.curve().genus()
        return not (self._u.is_one() and self._v.is_zero() and self._n == (g + 1) // 2)

    def __iter__(self) -> Iterator[Polynomial]:
        r"""
        TESTS:

        Indirect tests::

            sage: R.<x> = GF(7)[]
            sage: H = HyperellipticCurve(x^6 - x + 1)
            sage: J = H.jacobian()
            sage: P = J(x + 2, 5)
            sage: list(P)
            [x + 2, 5, 1]
            sage: tuple(P)
            (x + 2, 5, 1)
            sage: Q = J(x + 2, 5, 0)
            sage: list(Q)
            [x + 2, 5, 0]
            sage: tuple(Q)
            (x + 2, 5, 0)
            sage: P == Q
            False
        """
        yield from [self._u, self._v, self._n]

    def __hash__(self) -> int:
        r"""
        Compute the hash value of this element.

        TESTS::

            sage: R.<x> = GF(23)[]
            sage: H = HyperellipticCurve(x^6 + 1)
            sage: J = H.jacobian()
            sage: D1 = J(x + 6, 17, 0)
            sage: D2 = J(x + 6, 17, 1)
            sage: hash(D1) == hash(D2)
            False
        """
        return hash(tuple(self))

    def _add_(self, other) -> Self:
        r"""
        Return the sum of the two elements.

        Follows algorithm 3.7 of [Mireles2008]_.

        EXAMPLES::

            sage: R.<x> = QQ[]
            sage: H = HyperellipticCurve(x^5 - x^4 + x^2 - x, x^3 + 1)
            sage: J = Jacobian(H)
            sage: D1 = J(x^2 + x, 0)
            sage: D2 = J(x^2, -x)
            sage: D1 + D2
            (x^2 - 3/2*x + 1/2, -5/2*x + 1/2 : 0)
            sage: D3 = J(x - 1, -2)
            sage: D1 + D3
            (x^2 - 1/2*x, 5/4*x - 1 : 0)
        """
        # Collect data from HyperellipticCurve
        H = self.parent().curve()
        g = H.genus()

        # Extract out mumford coordinates
        u1, v1 = self.uv()
        u2, v2 = other.uv()

        # Extract out integers for weights
        n1, n2 = self._n, other._n

        # Step one: cantor composition of the two divisors
        u3, v3, n3 = self._parent.cantor_composition(u1, v1, n1, u2, v2, n2)

        # Step two: cantor reduction of the above to ensure
        # the degree of u is smaller than g + 1
        while u3.degree() > (g + 1):
            u3, v3, n3 = self._parent.cantor_reduction(u3, v3, n3)

        # Step three: compose and then reduce at infinity to ensure
        # unique representation of D
        while n3 < 0 or n3 > g - u3.degree():
            u3, v3, n3 = self._parent.cantor_compose_at_infinity(
                u3, v3, n3, plus=(n3 >= 0)
            )

        return self._parent(u3, v3, n3, check=False)

    def _neg_(self):
        r"""
        Return the divisor multiplied by -1.

        Follows algorithm 3.8 of [Mireles2008]_.

        TESTS::

            sage: R.<x> = GF(101)[]
            sage: H = HyperellipticCurve(x^5 + 1, x^3 + x)
            sage: assert H.is_split()
            sage: J = Jacobian(H)
            sage: [P0,P1] = H.points_at_infinity()
            sage: P = H.random_point()
            sage: Q = H.hyperelliptic_involution(P)
            sage: - J(P,P0) == J(Q,P1)
            True
        """
        # Collect data from HyperellipticCurve
        H = self.parent().curve()
        _, h = H.hyperelliptic_polynomials()
        g = H.genus()

        u0, v0 = self.uv()
        n0, m0 = self._n, self._m

        # Case for even genus
        if not (g % 2):
            v1 = (-h - v0) % u0
            u1 = u0
            n1 = m0
        # Odd genus, positive n0
        elif n0 > 0:
            v1 = (-h - v0) % u0
            u1 = u0
            # There is a typo for n1 in [GHM2008]_ and [Mireles2008]_.
            # This is corrected on page 617 of [Gal2018]_.
            n1 = m0 + 1
        else:
            # Composition at infinity always with plus=True.
            # want to "subtract" \infty_+ - \infty_-
            (u1, v1, n1) = self._parent.cantor_compose_at_infinity(
                u0, -h - v0, n0, plus=True
            )
            n1 = n1 - n0 + m0 + 1  # Shouldn't this always be 0? (See Alg 3.8 of [Mireles2008]_)
            assert n1 == 0

        return self._parent(u1, v1, n1, check=False)



"""
The C3 algorithm

The C3 algorithm is used as method resolution order for new style
classes in Python. The implementation here is used to order the list
of super categories of a category.

AUTHOR:

- Simon King (2011-11): initial version.
"""

# ****************************************************************************
#       Copyright (C) 2011 Simon King <simon.king@uni-jena.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************


cpdef list C3_algorithm(object start, str bases, str attribute, bint proper):
    """
    An implementation of the C3 algorithm.

    C3 is the algorithm used by Python to construct the method
    resolution order for new style classes involving multiple
    inheritance.

    After :issue:`11943` this implementation was used to compute the
    list of super categories of a category; see
    :meth:`~sage.categories.category.Category.all_super_categories`.
    The purpose is to ensure that list of super categories matches
    with the method resolution order of the parent or element classes
    of a category.

    Since :issue:`13589`, this implementation is superseded by that in
    :mod:`sage.misc.c3_controlled`, that puts the ``C3`` algorithm
    under control of some total order on categories.  This guarantees
    that ``C3`` always finds a consistent Method Resolution Order. For
    background, see :mod:`sage.misc.c3_controlled`.

    INPUT:

    - ``start`` -- an object; the returned list is built upon data
      provided by certain attributes of ``start``
    - ``bases`` -- string; the name of an attribute of ``start``
      providing a list of objects
    - ``attribute`` -- string; the name of an attribute of the
      objects provided in ``getattr(start,bases)``. That attribute is
      supposed to provide a list.

    ASSUMPTIONS:

    Our implementation of the algorithm only works on lists of
    objects that compare equal if and only if they are identical.

    OUTPUT:

    A list, the result of the C3 algorithm applied to the list
    ``[getattr(X,attribute) for X in getattr(start,bases)]``.

    EXAMPLES:

    We create a class for elements in a hierarchy that uses the ``C3``
    algorithm to compute, for each element, a linear extension of the
    elements above it::

    .. TODO:: Move back the __init__ at the beginning

        sage: from sage.misc.c3 import C3_algorithm
        sage: class HierarchyElement(UniqueRepresentation):
        ....:     @lazy_attribute
        ....:     def _all_bases(self):
        ....:         return C3_algorithm(self, '_bases', '_all_bases', False)
        ....:     def __repr__(self):
        ....:         return self._name
        ....:     def __init__(self, name, bases):
        ....:         self._name = name
        ....:         self._bases = list(bases)

    We construct a little hierarchy::

        sage: T = HierarchyElement("T", ())
        sage: X = HierarchyElement("X", (T,))
        sage: Y = HierarchyElement("Y", (T,))
        sage: A = HierarchyElement("A", (X, Y))
        sage: B = HierarchyElement("B", (Y, X))
        sage: Foo = HierarchyElement("Foo", (A, B))

    And inspect the linear extensions associated to each element::

        sage: T._all_bases
        [T]
        sage: X._all_bases
        [X, T]
        sage: Y._all_bases
        [Y, T]
        sage: A._all_bases
        [A, X, Y, T]
        sage: B._all_bases
        [B, Y, X, T]

    So far so good. However::

        sage: Foo._all_bases
        Traceback (most recent call last):
        ...
        ValueError: Cannot merge the items X, Y.

    The ``C3`` algorithm is not able to create a consistent linear
    extension. Indeed, its specifications impose that, if ``X`` and
    ``Y`` appear in a certain order in the linear extension for an
    element of the hierarchy, then they should appear in the same
    order for any lower element. This is clearly not possibly for
    ``Foo``, since ``A`` and ``B`` impose incompatible orders. If the
    above was a hierarchy of classes, Python would complain that it
    cannot calculate a consistent Method Resolution Order.

    TESTS:

    Regression test for bug #1 of :issue:`13501`::

        sage: class C(): pass
        sage: class F(): pass
        sage: class G(): pass
        sage: class B(C,F):    pass
        sage: class D(F,G):    pass
        sage: class E(F):      pass
        sage: class A(B,D,E):  pass
        sage: [cls.__name__ for cls in A.mro()]
        ['A', 'B', 'C', 'D', 'E', 'F', 'G', 'object']

        sage: C = HierarchyElement("C", ())
        sage: F = HierarchyElement("F", ())
        sage: G = HierarchyElement("G", ())
        sage: B = HierarchyElement("B", (C, F))
        sage: D = HierarchyElement("D", (F, G))
        sage: E = HierarchyElement("E", (F,))
        sage: A = HierarchyElement("A", (B, D, E))
        sage: A._all_bases
        [A, B, C, D, E, F, G]

    Regression test for bug #2 of :issue:`13501`. The following should
    fail since ``A`` asks for ``B`` to come before ``C``, where as
    ``B`` is a super class of ``C``::

        sage: class B(): pass
        sage: class C(B): pass
        sage: class A(B, C): pass
        Traceback (most recent call last):
        ...
        TypeError: ...Cannot create a consistent method resolution order (MRO)
        for bases ...

        sage: B = HierarchyElement("B", ())
        sage: C = HierarchyElement("C", (B,))
        sage: A = HierarchyElement("A", (B,C))
        sage: A._all_bases
        Traceback (most recent call last):
        ...
        ValueError: Cannot merge the items B, C, B.

    Since :issue:`11943`, the following consistency tests are part
    of the test suites of categories (except for hom categories)::

        sage: C = Category.join([HopfAlgebrasWithBasis(QQ), FiniteEnumeratedSets()])
        sage: C.parent_class.mro() == [x.parent_class for x in C.all_super_categories()]+[object]
        True
        sage: C.element_class.mro() == [x.element_class for x in C.all_super_categories()]+[object]
        True
    """
    cdef list out
    if proper:
        out = []
    else:
        out = [start]
    cdef list args = getattr(start, bases)
    if not args:
        return out
    # Data structure / invariants:
    # We will be working with the MRO's of the super objects
    # together with the list of bases of ``self``.
    # Each list is split between its head (in ``heads``) and tail (in
    # ``tails'') . Each tail is stored reversed, so that we can use a
    # cheap pop() in lieue of pop(0). A duplicate of the tail is
    # stored as a set in ``tailsets`` for cheap membership testing.
    # Since we actually want comparison by identity, not equality,
    # what we store is the set of memory locations of objects
    cdef object O, X
    cdef list tails = [getattr(obj, attribute) for obj in args]
    tails.append(args)
    tails = [list(reversed(tail)) for tail in tails]
    cdef list heads = [tail.pop() for tail in tails]
    cdef list tailsets = [set([<size_t><void *>O for O in tail])
                          for tail in tails]

    cdef int i, j, nbheads
    nbheads = len(heads)
    cdef bint next_item_found
    cdef list tail_list

    while nbheads:
        for i in range(nbheads):
            O = heads[i]
            # Does O appear in none of the tails?  ``all(O not in tail for tail in tailsets)``
            next_item_found = True
            for j in range(nbheads):
                if j == i:
                    continue
                if <size_t><void *>O in <set>tailsets[j]:
                    next_item_found = False
                    break
            if next_item_found:
                out.append(O)
                # Clear O from other heads, removing the line altogether
                # if the tail is already empty.
                # j goes down so that ``del heads[j]`` does not screw up the numbering
                for j in range(nbheads - 1, -1, -1):
                    if heads[j] is O:
                        tail_list = tails[j]
                        if tail_list:
                            X = tail_list.pop()
                            heads[j] = X
                            <set>tailsets[j].remove(<size_t><void *>X)
                        else:
                            del heads[j]
                            del tails[j]
                            del tailsets[j]
                            nbheads -= 1
                break
        if not next_item_found:
            # No head is available
            raise ValueError("Cannot merge the items {}.".format(', '.join([repr(head) for head in heads])))
    return out



r"""
An interface to Anders Buch's Littlewood-Richardson Calculator ``lrcalc``

The "Littlewood-Richardson Calculator" is a C library for fast
computation of Littlewood-Richardson (LR) coefficients and products of
Schubert polynomials. It handles single LR coefficients, products of
and coproducts of Schur functions, skew Schur functions, and
fusion products. All of the above are achieved by counting LR
(skew)-tableaux (also called Yamanouchi (skew)-tableaux) of
appropriate shape and content by iterating through them.
Additionally, ``lrcalc`` handles products of Schubert polynomials.

The web page of ``lrcalc`` is
`<http://sites.math.rutgers.edu/~asbuch/lrcalc/>`_.

The following describes the Sage interface to this library.

EXAMPLES::

    sage: import sage.libs.lrcalc.lrcalc as lrcalc

Compute a single Littlewood-Richardson coefficient::

    sage: lrcalc.lrcoef([3,2,1],[2,1],[2,1])
    2

Compute a product of Schur functions; return the coefficients in the
Schur expansion::

    sage: lrcalc.mult([2,1], [2,1])
    {[2, 2, 1, 1]: 1,
     [2, 2, 2]: 1,
     [3, 1, 1, 1]: 1,
     [3, 2, 1]: 2,
     [3, 3]: 1,
     [4, 1, 1]: 1,
     [4, 2]: 1}

Same product, but include only partitions with at most 3 rows.  This
corresponds to computing in the representation ring of `\mathfrak{gl}(3)`::

    sage: lrcalc.mult([2,1], [2,1], 3)
    {[2, 2, 2]: 1, [3, 2, 1]: 2, [3, 3]: 1, [4, 1, 1]: 1, [4, 2]: 1}

We can also compute the fusion product, here for `\mathfrak{sl}(3)`
and level 2::

    sage: lrcalc.mult([3,2,1], [3,2,1], 3,2)
    {[4, 4, 4]: 1, [5, 4, 3]: 1}

Compute the expansion of a skew Schur function::

    sage: lrcalc.skew([3,2,1],[2,1])
    {[1, 1, 1]: 1, [2, 1]: 2, [3]: 1}

Compute the coproduct of a Schur function::

    sage: lrcalc.coprod([3,2,1])
    {([1, 1, 1], [2, 1]): 1,
     ([2, 1], [2, 1]): 2,
     ([2, 1], [3]): 1,
     ([2, 1, 1], [1, 1]): 1,
     ([2, 1, 1], [2]): 1,
     ([2, 2], [1, 1]): 1,
     ([2, 2], [2]): 1,
     ([2, 2, 1], [1]): 1,
     ([3, 1], [1, 1]): 1,
     ([3, 1], [2]): 1,
     ([3, 1, 1], [1]): 1,
     ([3, 2], [1]): 1,
     ([3, 2, 1], []): 1}

Multiply two Schubert polynomials::

    sage: lrcalc.mult_schubert([4,2,1,3], [1,4,2,5,3])
    {[4, 5, 1, 3, 2]: 1,
     [5, 3, 1, 4, 2]: 1,
     [5, 4, 1, 2, 3]: 1,
     [6, 2, 1, 4, 3, 5]: 1}

Same product, but include only permutations of 5 elements in the result.
This corresponds to computing in the cohomology ring of `Fl(5)`::

    sage: lrcalc.mult_schubert([4,2,1,3], [1,4,2,5,3], 5)
    {[4, 5, 1, 3, 2]: 1, [5, 3, 1, 4, 2]: 1, [5, 4, 1, 2, 3]: 1}

List all Littlewood-Richardson tableaux of skew shape `\mu/\nu`; in
this example `\mu=[3,2,1]` and `\nu=[2,1]`. Specifying a third entry
`M' = ``maxrows`` restricts the alphabet to `\{1,2,\ldots,M\}`::

    sage: list(lrcalc.lrskew([3,2,1],[2,1]))
    [[[None, None, 1], [None, 1], [1]], [[None, None, 1], [None, 1], [2]],
    [[None, None, 1], [None, 2], [1]], [[None, None, 1], [None, 2], [3]]]

    sage: list(lrcalc.lrskew([3,2,1],[2,1],maxrows=2))
    [[[None, None, 1], [None, 1], [1]], [[None, None, 1], [None, 1], [2]],
     [[None, None, 1], [None, 2], [1]]]

.. TODO::

    Use this library in the :class:`SymmetricFunctions` code, to
    make it easy to apply it to linear combinations of Schur functions.

.. SEEALSO::

    - :func:`lrcoef`
    - :func:`mult`
    - :func:`coprod`
    - :func:`skew`
    - :func:`lrskew`
    - :func:`mult_schubert`

.. RUBRIC:: Underlying algorithmic in lrcalc

Here is some additional information regarding the main low-level
C-functions in `lrcalc`. Given two partitions ``outer`` and ``inner``
with ``inner`` contained in ``outer``, the function::

    skewtab *st_new(vector *outer, vector *inner, vector *conts, int maxrows)

constructs and returns the (lexicographically) first LR skew tableau
of shape ``outer / inner``. Further restrictions can be imposed using
``conts`` and ``maxrows``.

Namely, the integer ``maxrows`` is a bound on the integers that can be
put in the tableau.  The name is chosen because this will limit the
partitions in the output of :func:`skew` or :func:`mult` to partitions
with at most this number of rows.

The vector ``conts`` is the content of an empty tableau(!!). More
precisely, this vector is added to the usual content of a tableau
whenever the content is needed.  This affects which tableaux are
considered LR tableaux (see :func:`mult` below).  ``conts`` may also
be the ``NULL`` pointer, in which case nothing is added.

The other function::

    int *st_next(skewtab *st)

computes in place the (lexicographically) next skew tableau with the
same constraints, or returns 0 if ``st`` is the last one.

For a first example, see the :func:`skew` function code in the
``lrcalc`` source code. We want to compute a skew Schur function, so
create a skew LR tableau of the appropriate shape with ``st_new``
(with ``conts = NULL``), then iterate through all the LR tableaux with
``st_next()``. For each skew tableau, we use that ``st->conts`` is the
content of the skew tableau, find this shape in the ``res`` hash table
and add one to the value.

For a second example, see ``mult(vector *sh1, vector *sh2, maxrows)``.
Here we call ``st_new()`` with the shape ``sh1 / (0)`` and use ``sh2``
as the ``conts`` argument.  The effect of using ``sh2`` in this way is
that ``st_next`` will iterate through semistandard tableaux `T` of
shape ``sh1`` such that the following tableau::

         111111
         22222    <--- minimal tableau of shape sh2
         333
    *****
    **T**
    ****
    **

is a LR skew tableau, and ``st->conts`` contains the content of the
combined tableaux.

More generally, ``st_new(outer, inner, conts, maxrows)`` and
``st_next`` can be used to compute the Schur expansion of the product
``S_{outer/inner} * S_conts``, restricted to partitions with at most
``maxrows`` rows.

AUTHORS:

- Mike Hansen (2010): core of the interface

- Anne Schilling, Nicolas M. Thiéry, and Anders Buch (2011): fusion
  product, iterating through LR tableaux, finalization, documentation
"""
# ****************************************************************************
#  Copyright (C) 2010 Mike Hansen <mhansen@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.combinat.partition import _Partitions
from sage.combinat.permutation import Permutation
from sage.combinat.skew_tableau import SemistandardSkewTableaux
from sage.combinat.skew_partition import SkewPartition
from sage.rings.integer import Integer
import lrcalc


def _lrcalc_dict_to_sage(result) -> dict:
    r"""
    Translate from lrcalc output format to Sage expected format.

    TESTS::

        sage: from sage.libs.lrcalc.lrcalc import mult
        sage: mult([2,1],[3,2,1],3) # indirect doctest
        {[3, 3, 3]: 1, [4, 3, 2]: 2, [4, 4, 1]: 1, [5, 2, 2]: 1, [5, 3, 1]: 1}
    """
    return {_Partitions.element_class(_Partitions, [Integer(p) for p in la]):
            Integer(k) for la, k in result.items()}


def lrcoef_unsafe(outer, inner1, inner2):
    r"""
    Compute a single Littlewood-Richardson coefficient.

    Return the coefficient of ``outer`` in the product of the Schur
    functions indexed by ``inner1`` and ``inner2``.

    INPUT:

    - ``outer`` -- a partition (weakly decreasing list of nonnegative integers)
    - ``inner1`` -- a partition
    - ``inner2`` -- a partition

    .. WARNING::

       This function does not do any check on its input.  If you want
       to use a safer version, use :func:`lrcoef`.

    EXAMPLES::

        sage: from sage.libs.lrcalc.lrcalc import lrcoef_unsafe
        sage: lrcoef_unsafe([3,2,1], [2,1], [2,1])
        2
        sage: lrcoef_unsafe([3,3], [2,1], [2,1])
        1
        sage: lrcoef_unsafe([2,1,1,1,1], [2,1], [2,1])
        0
    """
    return Integer(lrcalc.lrcoef(outer, inner1, inner2))


def lrcoef(outer, inner1, inner2):
    """
    Compute a single Littlewood-Richardson coefficient.

    Return the coefficient of ``outer`` in the product of the Schur
    functions indexed by ``inner1`` and ``inner2``.

    INPUT:

    - ``outer`` -- a partition (weakly decreasing list of nonnegative integers)
    - ``inner1`` -- a partition
    - ``inner2`` -- a partition

    .. NOTE::

       This function converts its inputs into :func:`Partition`'s.  If
       you don't need these checks and your inputs are valid, then you
       can use :func:`lrcoef_unsafe`.

    EXAMPLES::

        sage: from sage.libs.lrcalc.lrcalc import lrcoef
        sage: lrcoef([3,2,1], [2,1], [2,1])
        2
        sage: lrcoef([3,3], [2,1], [2,1])
        1
        sage: lrcoef([2,1,1,1,1], [2,1], [2,1])
        0
    """
    return lrcoef_unsafe(_Partitions(outer), _Partitions(inner1), _Partitions(inner2))


def mult(part1, part2, maxrows=None, level=None, quantum=None) -> dict:
    r"""
    Compute a product of two Schur functions.

    Return the product of the Schur functions indexed by the
    partitions ``part1`` and ``part2``.

    INPUT:

    - ``part1`` -- a partition
    - ``part2`` -- a partition
    - ``maxrows`` -- integer (optional)
    - ``level`` -- integer (optional)
    - ``quantum`` -- an element of a ring (optional)

    If ``maxrows`` is specified, then only partitions with at most
    this number of rows are included in the result.

    If both ``maxrows`` and ``level`` are specified, then the function
    calculates the fusion product for `\mathfrak{sl}(\mathrm{maxrows})`
    of the given level.

    If ``quantum`` is set, then this returns the product in the quantum
    cohomology ring of the Grassmannian. In particular, both ``maxrows``
    and ``level`` need to be specified.

    EXAMPLES::

        sage: from sage.libs.lrcalc.lrcalc import mult
        sage: mult([2],[])
        {[2]: 1}
        sage: sorted(mult([2],[2]).items())
        [([2, 2], 1), ([3, 1], 1), ([4], 1)]
        sage: sorted(mult([2,1],[2,1]).items())
        [([2, 2, 1, 1], 1), ([2, 2, 2], 1), ([3, 1, 1, 1], 1),
         ([3, 2, 1], 2), ([3, 3], 1), ([4, 1, 1], 1), ([4, 2], 1)]
        sage: sorted(mult([2,1],[2,1],maxrows=2).items())
        [([3, 3], 1), ([4, 2], 1)]
        sage: mult([2,1],[3,2,1],3)
        {[3, 3, 3]: 1, [4, 3, 2]: 2, [4, 4, 1]: 1, [5, 2, 2]: 1, [5, 3, 1]: 1}
        sage: mult([2,1],[2,1],3,3)
        {[2, 2, 2]: 1, [3, 2, 1]: 2, [3, 3]: 1, [4, 1, 1]: 1}
        sage: mult([2,1],[2,1],None,3)
        Traceback (most recent call last):
        ...
        ValueError: maxrows needs to be specified if you specify the level

     The quantum product::

        sage: q = polygen(QQ, 'q')
        sage: sorted(mult([1],[2,1], 2, 2, quantum=q).items())
        [([], q), ([2, 2], 1)]
        sage: sorted(mult([2,1],[2,1], 2, 2, quantum=q).items())
        [([1, 1], q), ([2], q)]

        sage: mult([2,1],[2,1], quantum=q)
        Traceback (most recent call last):
        ...
        ValueError: missing parameters maxrows or level
    """
    if maxrows is None and level is not None:
        raise ValueError('maxrows needs to be specified if you specify'
                         ' the level')
    if quantum is not None and (level is None or maxrows is None):
        raise ValueError('missing parameters maxrows or level')

    if quantum is None:
        if level is not None:
            return _lrcalc_dict_to_sage(lrcalc.mult_fusion(part1, part2, maxrows, level))
        if maxrows is None:
            maxrows = -1
        return _lrcalc_dict_to_sage(lrcalc.mult(part1, part2, maxrows))

    # Otherwise do quantum multiplication
    result = lrcalc.mult_quantum(part1, part2, maxrows, level, degrees=True)
    P = quantum.parent()
    output = {}
    for i, k in result.items():
        la = _Partitions(i[0])
        output[la] = output.get(la, P.zero()) + k * quantum**(i[1])
    return output


def skew(outer, inner, maxrows=-1) -> dict:
    """
    Compute the Schur expansion of a skew Schur function.

    Return a linear combination of partitions representing the Schur
    function of the skew Young diagram ``outer / inner``, consisting
    of boxes in the partition ``outer`` that are not in ``inner``.

    INPUT:

    - ``outer`` -- a partition
    - ``inner`` -- a partition
    - ``maxrows`` -- integer or ``None``

    If ``maxrows`` is specified, then only partitions with at most
    this number of rows are included in the result.

    EXAMPLES::

        sage: from sage.libs.lrcalc.lrcalc import skew
        sage: sorted(skew([2,1],[1]).items())
        [([1, 1], 1), ([2], 1)]
    """
    return _lrcalc_dict_to_sage(lrcalc.skew(outer, inner, maxrows))


def coprod(part, all=0) -> dict:
    """
    Compute the coproduct of a Schur function.

    Return a linear combination of pairs of partitions representing
    the coproduct of the Schur function given by the partition
    ``part``.

    INPUT:

    - ``part`` -- a partition
    - ``all`` -- integer

    If ``all`` is nonzero then all terms are included in the result.
    If ``all`` is zero, then only pairs of partitions ``(part1,
    part2)`` for which the weight of ``part1`` is greater than or
    equal to the weight of ``part2`` are included; the rest of the
    coefficients are redundant because Littlewood-Richardson
    coefficients are symmetric.

    EXAMPLES::

        sage: from sage.libs.lrcalc.lrcalc import coprod
        sage: sorted(coprod([2,1]).items())
        [(([1, 1], [1]), 1), (([2], [1]), 1), (([2, 1], []), 1)]
    """
    result = lrcalc.coprod(part, all)
    return {tuple([_Partitions.element_class(_Partitions,
                                             [Integer(p) for p in mu])
                   for mu in la]): Integer(k)
            for la, k in result.items()}


def mult_schubert(w1, w2, rank=0) -> dict:
    r"""
    Compute a product of two Schubert polynomials.

    Return a linear combination of permutations representing the
    product of the Schubert polynomials indexed by the permutations
    ``w1`` and ``w2``.

    INPUT:

    - ``w1`` -- a permutation
    - ``w2`` -- a permutation
    - ``rank`` -- integer

    If ``rank`` is nonzero, then only permutations from the symmetric
    group `S(\mathrm{rank})` are included in the result.

    EXAMPLES::

        sage: from sage.libs.lrcalc.lrcalc import mult_schubert
        sage: result = mult_schubert([3, 1, 5, 2, 4], [3, 5, 2, 1, 4])
        sage: sorted(result.items())
        [([5, 4, 6, 1, 2, 3], 1), ([5, 6, 3, 1, 2, 4], 1),
         ([5, 7, 2, 1, 3, 4, 6], 1), ([6, 3, 5, 1, 2, 4], 1),
         ([6, 4, 3, 1, 2, 5], 1), ([6, 5, 2, 1, 3, 4], 1),
         ([7, 3, 4, 1, 2, 5, 6], 1), ([7, 4, 2, 1, 3, 5, 6], 1)]
    """
    result = lrcalc.schubmult(w1, w2, rank)
    return {Permutation(list(la)): Integer(k) for la, k in result.items()}


def lrskew(outer, inner, weight=None, maxrows=-1):
    r"""
    Iterate over the skew LR tableaux of shape ``outer / inner``.

    INPUT:

    - ``outer`` -- a partition
    - ``inner`` -- a partition
    - ``weight`` -- a partition (optional)
    - ``maxrows`` -- positive integer (optional)

    OUTPUT: an iterator of :class:`SkewTableau`

    Specifying ``maxrows`` = `M` restricts the alphabet to `\{1,2,\ldots,M\}`.

    Specifying ``weight`` returns only those tableaux of given content/weight.

    EXAMPLES::

        sage: from sage.libs.lrcalc.lrcalc import lrskew
        sage: for st in lrskew([3,2,1],[2]):
        ....:     st.pp()
        .  .  1
        1  1
        2
        .  .  1
        1  2
        2
        .  .  1
        1  2
        3

        sage: for st in lrskew([3,2,1],[2], maxrows=2):
        ....:     st.pp()
        .  .  1
        1  1
        2
        .  .  1
        1  2
        2

        sage: list(lrskew([3,2,1],[2], weight=[3,1]))
        [[[None, None, 1], [1, 1], [2]]]

    TESTS::

        sage: from sage.libs.lrcalc.lrcalc import lrskew
        sage: list(lrskew([3,2,1],[2], weight=[]))
        []
        sage: list(lrskew([3,2,1],[2], weight=[0]))
        []
        sage: list(lrskew([3,2,1],[3,2,1], weight=[]))
        [[[None, None, None], [None, None], [None]]]
        sage: list(lrskew([3,2,1],[3,2,1], weight=[0]))
        [[[None, None, None], [None, None], [None]]]
        sage: list(lrskew([3,2,1],[3,2,1], weight=[1]))
        []
    """
    iterator = lrcalc.lr_iterator(outer, inner, maxrows)
    shape = SkewPartition([outer, inner])

    if weight is None:
        ST = SemistandardSkewTableaux(shape)
        for data in iterator:
            yield ST.from_shape_and_word(shape, [i + 1 for i in data])
    else:
        wt = _Partitions(weight)
        ST = SemistandardSkewTableaux(shape, wt)
        m = len(wt)
        for data in iterator:
            w = [0] * m
            for j in data:
                if j >= m:
                    # We know they are not equal, so make the check below quick
                    w = None
                    break
                w[j] += 1
            if w == wt:
                yield ST.from_shape_and_word(shape, [i + 1 for i in data])



from sage.structure.factorization import Factorization

from sage.structure.sequence import Sequence, seq

from sage.structure.unique_representation import UniqueRepresentation

from sage.structure.sage_object import SageObject

from sage.structure.element import (
    canonical_coercion,
    coercion_model,
    get_coercion_model,
    coercion_traceback,
    parent
)

from sage.structure.parent import Parent

from sage.structure.parent_gens import localvars

from sage.structure.proof import all as proof

from sage.structure.mutability import Mutability

from sage.structure.element_wrapper import ElementWrapper

from sage.misc.lazy_import import lazy_import
lazy_import('sage.structure.formal_sum', ['FormalSums', 'FormalSum'])
del lazy_import



"""
Basic Statistics

This file contains basic descriptive functions. Included are the mean,
median, mode, moving average, standard deviation, and the variance.
When calling a function on data, there are checks for functions already
defined for that data type.

The :func:`mean` function returns the arithmetic mean (the sum of all the members
of a list, divided by the number of members). Further revisions may include
the geometric and harmonic mean. The :func:`median` function returns the number
separating the higher half of a sample from the lower half. The :func:`mode`
returns the most common occurring member of a sample, plus the number of times
it occurs. If entries occur equally common, the smallest of a list of the most
common  entries is returned. The :func:`moving_average` is a finite impulse
response filter, creating a series of averages using a user-defined number of
subsets of the full data set. The :func:`std` and the :func:`variance` return a
measurement of how far data points tend to be from the arithmetic mean.

Functions are available in the namespace :mod:`stats`, i.e. you can use them by
typing ``stats.mean``, ``stats.median``, etc.

REMARK: If all the data you are working with are floating point
numbers, you may find :class:`stats.TimeSeries` helpful, since it is
extremely fast and offers many of the same descriptive statistics as
in the module.

AUTHOR:

- Andrew Hou (11/06/2009)
"""
# ***********************************************************************
#          Copyright (C) 2009, Andrew Hou <amhou@uw.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#            The full text of the GPL is available at:
#                  https://www.gnu.org/licenses/
######################################################################

from sage.misc.functional import sqrt
from sage.misc.lazy_import import lazy_import
from sage.misc.superseded import deprecation
from sage.rings.integer_ring import ZZ

lazy_import("sage.symbolic.constants", "NaN")
lazy_import("numpy", "ndarray", as_="numpy_ndarray")
lazy_import("sage.stats.time_series", "TimeSeries")


def mean(v):
    """
    Return the mean of the elements of `v`.

    We define the mean of the empty list to be the (symbolic) NaN,
    following the convention of MATLAB, Scipy, and R.

    This function is deprecated.  Use :func:`numpy.mean` or :func:`numpy.nanmean`
    instead.

    INPUT:

    - ``v`` -- list of numbers

    OUTPUT: a number

    EXAMPLES::

        sage: mean([pi, e])                                                             # needs sage.symbolic
        doctest:warning...
        DeprecationWarning: sage.stats.basic_stats.mean is deprecated;
        use numpy.mean or numpy.nanmean instead
        See https://github.com/sagemath/sage/issues/29662 for details.
        1/2*pi + 1/2*e
        sage: mean([])                                                                  # needs sage.symbolic
        NaN
        sage: mean([I, sqrt(2), 3/5])                                                   # needs sage.symbolic
        1/3*sqrt(2) + 1/3*I + 1/5
        sage: mean([RIF(1.0103,1.0103), RIF(2)])
        1.5051500000000000?
        sage: mean(range(4))
        3/2
        sage: v = stats.TimeSeries([1..100])                                            # needs numpy
        sage: mean(v)                                                                   # needs numpy
        50.5
    """
    deprecation(29662, 'sage.stats.basic_stats.mean is deprecated; use numpy.mean or numpy.nanmean instead')
    if hasattr(v, 'mean'):
        return v.mean()
    if not v:
        return NaN
    s = sum(v)
    if isinstance(s, int):
        # python integers are stupid.
        return s / ZZ(len(v))
    return s / len(v)


def mode(v):
    """
    Return the mode of `v`.

    The mode is the list of the most frequently occurring
    elements in `v`. If `n` is the most times that any element occurs
    in `v`, then the mode is the list of elements of `v` that
    occur `n` times. The list is sorted if possible.

    This function is deprecated.  Use :func:`scipy:scipy.stats.mode` or
    :func:`statistics.mode` instead.

    .. NOTE::

        The elements of `v` must be hashable.

    INPUT:

    - ``v`` -- list

    OUTPUT: list (sorted if possible)

    EXAMPLES::

        sage: v = [1,2,4,1,6,2,6,7,1]
        sage: mode(v)
        doctest:warning...
        DeprecationWarning: sage.stats.basic_stats.mode is deprecated;
        use scipy.stats.mode or statistics.mode instead
        See https://github.com/sagemath/sage/issues/29662 for details.
        [1]
        sage: v.count(1)
        3
        sage: mode([])
        []

        sage: mode([1,2,3,4,5])
        [1, 2, 3, 4, 5]
        sage: mode([3,1,2,1,2,3])
        [1, 2, 3]
        sage: mode([0, 2, 7, 7, 13, 20, 2, 13])
        [2, 7, 13]

        sage: mode(['sage', 'four', 'I', 'three', 'sage', 'pi'])
        ['sage']

        sage: class MyClass:
        ....:   def mode(self):
        ....:       return [1]
        sage: stats.mode(MyClass())
        [1]
    """
    deprecation(29662, 'sage.stats.basic_stats.mode is deprecated; use scipy.stats.mode or statistics.mode instead')

    if hasattr(v, 'mode'):
        return v.mode()

    if not v:
        return v

    freq = {}
    for i in v:
        if i in freq:
            freq[i] += 1
        else:
            freq[i] = 1

    n = max(freq.values())
    try:
        return sorted(u for u, f in freq.items() if f == n)
    except TypeError:
        return [u for u, f in freq.items() if f == n]


def std(v, bias=False):
    """
    Return the standard deviation of the elements of `v`.

    We define the standard deviation of the empty list to be NaN,
    following the convention of MATLAB, Scipy, and R.

    This function is deprecated.  Use :func:`numpy.std` or :func:`numpy.nanstd`
    instead.

    INPUT:

    - ``v`` -- list of numbers

    - ``bias`` -- boolean (default: ``False``); if ``False``, divide by
      ``len(v) - 1`` instead of ``len(v)`` to give a less biased
      estimator (sample) for the standard deviation.

    OUTPUT: a number

    EXAMPLES::

        sage: # needs sage.symbolic
        sage: std([1..6], bias=True)
        doctest:warning...
        DeprecationWarning: sage.stats.basic_stats.std is deprecated;
        use numpy.std or numpy.nanstd instead
        See https://github.com/sagemath/sage/issues/29662 for details.
        doctest:warning...
        DeprecationWarning: sage.stats.basic_stats.variance is deprecated;
        use numpy.var or numpy.nanvar instead
        See https://github.com/sagemath/sage/issues/29662 for details.
        doctest:warning...
        DeprecationWarning: sage.stats.basic_stats.mean is deprecated;
        use numpy.mean or numpy.nanmean instead
        See https://github.com/sagemath/sage/issues/29662 for details.
        1/2*sqrt(35/3)
        sage: std([1..6], bias=False)
        sqrt(7/2)
        sage: std([e, pi])
        sqrt(1/2)*abs(pi - e)
        sage: std([])
        NaN
        sage: std([I, sqrt(2), 3/5])
        1/15*sqrt(1/2)*sqrt((10*sqrt(2) - 5*I - 3)^2
        + (5*sqrt(2) - 10*I + 3)^2 + (5*sqrt(2) + 5*I - 6)^2)
        sage: std([RIF(1.0103, 1.0103), RIF(2)])
        0.6998235813403261?

        sage: # needs numpy
        sage: import numpy
        sage: if int(numpy.version.short_version[0]) > 1:
        ....:     _ = numpy.set_printoptions(legacy="1.25")
        sage: x = numpy.array([1,2,3,4,5])
        sage: std(x, bias=False)
        1.5811388300841898
        sage: x = stats.TimeSeries([1..100])
        sage: std(x)
        29.011491975882016

    TESTS::

        sage: data = [random() for i in [1 .. 20]]
        sage: std(data)  # random
        0.29487771726609185
    """
    deprecation(29662, 'sage.stats.basic_stats.std is deprecated; use numpy.std or numpy.nanstd instead')

    # NOTE: in R bias = False by default, and in Scipy bias=True by
    # default, and R is more popular.

    if hasattr(v, 'standard_deviation'):
        return v.standard_deviation(bias=bias)

    if isinstance(v, numpy_ndarray):
        # accounts for numpy arrays
        if bias:
            return v.std()
        return v.std(ddof=1)

    if not v:
        # standard deviation of empty set defined as NaN
        return NaN

    return sqrt(variance(v, bias=bias))


def variance(v, bias=False):
    """
    Return the variance of the elements of `v`.

    We define the variance of the empty list to be NaN,
    following the convention of MATLAB, Scipy, and R.

    This function is deprecated.  Use :func:`numpy.var` or :func:`numpy.nanvar`
    instead.

    INPUT:

    - ``v`` -- list of numbers

    - ``bias`` -- boolean (default: ``False``); if ``False``, divide by
      ``len(v) - 1`` instead of ``len(v)`` to give a less biased
      estimator (sample) for the standard deviation.

    OUTPUT: a number

    EXAMPLES::

        sage: variance([1..6])
        doctest:warning...
        DeprecationWarning: sage.stats.basic_stats.variance is deprecated;
        use numpy.var or numpy.nanvar instead
        See https://github.com/sagemath/sage/issues/29662 for details.
        7/2
        sage: variance([1..6], bias=True)
        35/12
        sage: variance([e, pi])                                                         # needs sage.symbolic
        1/2*(pi - e)^2
        sage: variance([])
        NaN
        sage: variance([I, sqrt(2), 3/5])                                               # needs sage.symbolic
        1/450*(10*sqrt(2) - 5*I - 3)^2 + 1/450*(5*sqrt(2) - 10*I + 3)^2
        + 1/450*(5*sqrt(2) + 5*I - 6)^2
        sage: variance([RIF(1.0103, 1.0103), RIF(2)])
        0.4897530450000000?
        sage: import numpy                                                              # needs numpy
        sage: if int(numpy.version.short_version[0]) > 1:                               # needs numpy
        ....:     _ = numpy.set_printoptions(legacy="1.25")                                 # needs numpy
        sage: x = numpy.array([1,2,3,4,5])                                              # needs numpy
        sage: variance(x, bias=False)                                                   # needs numpy
        2.5
        sage: x = stats.TimeSeries([1..100])
        sage: variance(x)
        841.6666666666666
        sage: variance(x, bias=True)
        833.25
        sage: class MyClass:
        ....:   def variance(self, bias=False):
        ....:      return 1
        sage: stats.variance(MyClass())
        1
        sage: class SillyPythonList:
        ....:   def __init__(self):
        ....:       self.__list = [2, 4]
        ....:   def __len__(self):
        ....:       return len(self.__list)
        ....:   def __iter__(self):
        ....:       return self.__list.__iter__()
        ....:   def mean(self):
        ....:       return 3
        sage: R = SillyPythonList()
        sage: variance(R)
        2
        sage: variance(R, bias=True)
        1

    TESTS:

    The performance issue from :issue:`10019` is solved::

        sage: variance([1] * 2^18)
        0
    """
    deprecation(29662, 'sage.stats.basic_stats.variance is deprecated; use numpy.var or numpy.nanvar instead')

    if hasattr(v, 'variance'):
        return v.variance(bias=bias)

    x = 0
    if isinstance(v, numpy_ndarray):
        # accounts for numpy arrays
        if bias:
            return v.var()
        return v.var(ddof=1)
    if not v:
        # variance of empty set defined as NaN
        return NaN

    mu = mean(v)
    for vi in v:
        x += (vi - mu)**2
    if bias:
        # population variance
        if isinstance(x, int):
            return x / ZZ(len(v))
        return x / len(v)
    # sample variance
    if isinstance(x, int):
        return x / ZZ(len(v)-1)
    return x / (len(v)-1)


def median(v):
    """
    Return the median (middle value) of the elements of `v`.

    If `v` is empty, we define the median to be NaN, which is
    consistent with NumPy (note that R returns NULL).
    If `v` is comprised of strings, :exc:`TypeError` occurs.
    For elements other than numbers, the median is a result of :func:`sorted`.

    This function is deprecated.  Use :func:`numpy.median` or :func:`numpy.nanmedian`
    instead.

    INPUT:

    - ``v`` -- list

    OUTPUT: median element of `v`

    EXAMPLES::

        sage: median([1,2,3,4,5])
        doctest:warning...
        DeprecationWarning: sage.stats.basic_stats.median is deprecated;
        use numpy.median or numpy.nanmedian instead
        See https://github.com/sagemath/sage/issues/29662 for details.
        3
        sage: median([e, pi])                                                           # needs sage.symbolic
        1/2*pi + 1/2*e
        sage: median(['sage', 'linux', 'python'])
        'python'
        sage: median([])                                                                # needs sage.symbolic
        NaN
        sage: class MyClass:
        ....:    def median(self):
        ....:       return 1
        sage: stats.median(MyClass())
        1
    """
    deprecation(29662, 'sage.stats.basic_stats.median is deprecated; use numpy.median or numpy.nanmedian instead')

    if hasattr(v, 'median'):
        return v.median()

    if not v:
        # Median of empty set defined as NaN
        return NaN
    values = sorted(v)
    if len(values) % 2:
        return values[((len(values))+1)//2-1]
    lower = values[(len(values)+1)//2-1]
    upper = values[len(values)//2]
    return (lower + upper) / ZZ(2)


def moving_average(v, n):
    """
    Return the moving average of a list `v`.

    The moving average of a list is often used to smooth out noisy data.

    If `v` is empty, we define the entries of the moving average to be NaN.

    This method is deprecated.  Use :meth:`pandas.Series.rolling` instead.

    INPUT:

    - ``v`` -- list

    - ``n`` -- the number of values used in computing each average

    OUTPUT: list of length ``len(v)-n+1``, since we do not fabric any values

    EXAMPLES::

        sage: moving_average([1..10], 1)
        doctest:warning...
        DeprecationWarning: sage.stats.basic_stats.moving_average is deprecated;
        use pandas.Series.rolling instead
        See https://github.com/sagemath/sage/issues/29662 for details.
        [1, 2, 3, 4, 5, 6, 7, 8, 9, 10]
        sage: moving_average([1..10], 4)
        [5/2, 7/2, 9/2, 11/2, 13/2, 15/2, 17/2]
        sage: moving_average([], 1)
        []
        sage: moving_average([pi, e, I, sqrt(2), 3/5], 2)                               # needs sage.symbolic
        [1/2*pi + 1/2*e, 1/2*e + 1/2*I, 1/2*sqrt(2) + 1/2*I,
         1/2*sqrt(2) + 3/10]

    We check if the input is a time series, and if so use the
    optimized :meth:`simple_moving_average` method, but with (slightly
    different) meaning as defined above (the point is that the
    :meth:`simple_moving_average` on time series returns `n` values::

        sage: a = stats.TimeSeries([1..10])                                             # needs numpy
        sage: stats.moving_average(a, 3)                                                # needs numpy
        [2.0000, 3.0000, 4.0000, 5.0000, 6.0000, 7.0000, 8.0000, 9.0000]
        sage: stats.moving_average(list(a), 3)                                          # needs numpy
        [2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0]
    """
    deprecation(29662, 'sage.stats.basic_stats.moving_average is deprecated; use pandas.Series.rolling instead')

    if not v:
        return v
    if isinstance(v, TimeSeries):
        return v.simple_moving_average(n)[n - 1:]
    n = int(n)
    if n <= 0:
        raise ValueError("n must be positive")
    nn = ZZ(n)
    s = sum(v[:n])
    ans = [s / nn]
    for i in range(n, len(v)):
        # add in the i-th value in v to our running sum,
        # and remove the value n places back.
        s += v[i] - v[i - n]
        ans.append(s / nn)
    return ans



# cython: wraparound=False, boundscheck=False
"""
Braid orbit

Cython function to compute the orbit of the braid moves on a reduced word.
"""
from cysignals.signals cimport sig_check


cpdef set BraidOrbit(list word, list rels):
    r"""
    Return the orbit of ``word`` by all replacements given by ``rels``.

    INPUT:

    - ``word`` -- list of integers

    - ``rels`` -- list of pairs ``(A, B)``, where ``A`` and ``B`` are
      lists of integers the same length

    EXAMPLES::

        sage: from sage.combinat.root_system.braid_orbit import BraidOrbit
        sage: word = [1,2,1,3,2,1]
        sage: rels = [[[2, 1, 2], [1, 2, 1]], [[3, 1], [1, 3]], [[3, 2, 3], [2, 3, 2]]]
        sage: sorted(BraidOrbit(word, rels))
        [(1, 2, 1, 3, 2, 1),
         (1, 2, 3, 1, 2, 1),
         (1, 2, 3, 2, 1, 2),
         (1, 3, 2, 1, 3, 2),
         (1, 3, 2, 3, 1, 2),
         (2, 1, 2, 3, 2, 1),
         (2, 1, 3, 2, 1, 3),
         (2, 1, 3, 2, 3, 1),
         (2, 3, 1, 2, 1, 3),
         (2, 3, 1, 2, 3, 1),
         (2, 3, 2, 1, 2, 3),
         (3, 1, 2, 1, 3, 2),
         (3, 1, 2, 3, 1, 2),
         (3, 2, 1, 2, 3, 2),
         (3, 2, 1, 3, 2, 3),
         (3, 2, 3, 1, 2, 3)]
        sage: len(_)
        16
    """
    cdef int i, l, rel_l, loop_ind, list_len
    cdef tuple left, right, test_word, new_word
    cdef list rel

    l = len(word)
    cdef set words = {tuple(word)}
    cdef list test_words = [tuple(word)]

    rels = rels + [[b, a] for a, b in rels]
    rels = [[tuple(a), tuple(b), len(a)] for a, b in rels]

    loop_ind = 0
    list_len = 1
    while loop_ind < list_len:
        sig_check()
        test_word = <tuple> test_words[loop_ind]
        loop_ind += 1
        for rel in rels:
            left = <tuple> rel[0]
            right = <tuple> rel[1]
            rel_l = <int> rel[2]
            for i in range(l-rel_l+1):
                if pattern_match(test_word, i, left, rel_l):
                    new_word = test_word[:i] + right + test_word[i+rel_l:]
                    if new_word not in words:
                        words.add(new_word)
                        test_words.append(new_word)
                        list_len += 1
    return words


cpdef bint is_fully_commutative(list word, list rels) noexcept:
    r"""
    Check if the braid orbit of ``word`` is using a braid relation.

    INPUT:

    - ``word`` -- list of integers

    - ``rels`` -- list of pairs ``(A, B)``, where ``A`` and ``B`` are
      lists of integers the same length

    EXAMPLES::

        sage: from sage.combinat.root_system.braid_orbit import is_fully_commutative
        sage: rels = [[[2, 1, 2], [1, 2, 1]], [[3, 1], [1, 3]], [[3, 2, 3], [2, 3, 2]]]
        sage: word = [1,2,1,3,2,1]
        sage: is_fully_commutative(word, rels)
        False
        sage: word = [1,2,3]
        sage: is_fully_commutative(word, rels)
        True
    """
    cdef int i, l, rel_l, loop_ind, list_len
    cdef tuple left, right, test_word, new_word
    cdef list rel

    l = len(word)
    cdef set words = {tuple(word)}
    cdef list test_words = [tuple(word)]

    rels = rels + [[b, a] for a, b in rels]
    rels = [[tuple(a), tuple(b), len(a)] for a, b in rels]

    loop_ind = 0
    list_len = 1
    while loop_ind < list_len:
        sig_check()
        test_word = <tuple> test_words[loop_ind]
        loop_ind += 1
        for rel in rels:
            left = <tuple> rel[0]
            right = <tuple> rel[1]
            rel_l = <int> rel[2]
            for i in range(l-rel_l+1):
                if pattern_match(test_word, i, left, rel_l):
                    if rel_l > 2:
                        return False
                    new_word = test_word[:i] + right + test_word[i+rel_l:]
                    if new_word not in words:
                        words.add(new_word)
                        test_words.append(new_word)
                        list_len += 1
    return True


cdef inline bint pattern_match(tuple L, int i, tuple X, int l) noexcept:
    r"""
    Return ``True`` if ``L[i:i+l] == X``.

    Assumes that ``i`` is the length of ``L`` and ``l`` is the
    length of ``X``.
    """
    cdef int ind
    for ind in range(l):
        if <int> L[i+ind] != <int> X[ind]:
            return False
    return True



"""
Convert PARI objects to/from Flint objects

Utility function to convert PARI ``GEN``s to/from flint types.

AUTHORS:

- Luca De Feo (2016-09-06): Separate Sage-specific components from
  generic C-interface in ``Pari`` (:issue:`20241`)
"""

# ****************************************************************************
#       Copyright (C) 2016 Luca De Feo <luca.defeo@polytechnique.edu>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from cysignals.signals cimport sig_on

from sage.libs.flint.fmpz cimport COEFF_IS_MPZ, COEFF_TO_PTR, fmpz_is_one
from sage.libs.flint.fmpq cimport fmpq_numref, fmpq_denref
from sage.libs.flint.fmpz_mat cimport fmpz_mat_nrows, fmpz_mat_ncols, fmpz_mat_entry
from sage.libs.flint.fmpq_mat cimport fmpq_mat_nrows, fmpq_mat_ncols, fmpq_mat_entry

from cypari2.paridecl cimport *
from cypari2.stack cimport new_gen
from sage.libs.pari.convert_gmp cimport _new_GEN_from_mpz_t


cdef inline GEN _new_GEN_from_fmpz_t(fmpz_t value) noexcept:
    r"""
    Create a new PARI ``t_INT`` from a ``fmpz_t``.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    if COEFF_IS_MPZ(value[0]):
        return _new_GEN_from_mpz_t(COEFF_TO_PTR(value[0]))
    else:
        return stoi(value[0])


cdef inline GEN _new_GEN_from_fmpq_t(fmpq_t value) noexcept:
    r"""
    Create a new PARI ``t_RAT`` from a ``fmpq_t``.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    cdef GEN num = _new_GEN_from_fmpz_t(fmpq_numref(value))
    if fmpz_is_one(fmpq_denref(value)):
        return num
    cdef GEN denom = _new_GEN_from_fmpz_t(fmpq_denref(value))
    return mkfrac(num, denom)


cdef GEN _new_GEN_from_fmpz_mat_t(fmpz_mat_t B) noexcept:
    r"""
    Create a new PARI ``t_MAT`` with ``nr`` rows and ``nc`` columns
    from a ``fmpz_mat_t``.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    cdef GEN x
    cdef GEN A = zeromatcopy(fmpz_mat_nrows(B), fmpz_mat_ncols(B))
    cdef Py_ssize_t i, j
    for i in range(fmpz_mat_nrows(B)):
        for j in range(fmpz_mat_ncols(B)):
            x = _new_GEN_from_fmpz_t(fmpz_mat_entry(B, i, j))
            set_gcoeff(A, i+1, j+1, x)  # A[i+1, j+1] = x (using 1-based indexing)
    return A


cdef GEN _new_GEN_from_fmpq_mat_t(fmpq_mat_t B) noexcept:
    cdef GEN x
    cdef GEN A = zeromatcopy(fmpq_mat_nrows(B), fmpq_mat_ncols(B))
    cdef Py_ssize_t i, j
    for i in range(fmpq_mat_nrows(B)):
        for j in range(fmpq_mat_ncols(B)):
            x = _new_GEN_from_fmpq_t(fmpq_mat_entry(B, i, j))
            set_gcoeff(A, i+1, j+1, x)  # A[i+1, j+1] = x (using 1-based indexing)
    return A

cdef GEN _new_GEN_from_fmpz_mat_t_rotate90(fmpz_mat_t B) noexcept:
    r"""
    Create a new PARI ``t_MAT`` with ``nr`` rows and ``nc`` columns
    from a ``fmpz_mat_t`` and rotate the matrix 90 degrees
    counterclockwise.  So the resulting matrix will have ``nc`` rows
    and ``nr`` columns.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    cdef GEN x
    cdef GEN A = zeromatcopy(fmpz_mat_ncols(B), fmpz_mat_nrows(B))
    cdef Py_ssize_t i, j
    for i in range(fmpz_mat_nrows(B)):
        for j in range(fmpz_mat_ncols(B)):
            x = _new_GEN_from_fmpz_t(fmpz_mat_entry(B, i, fmpz_mat_ncols(B) - j - 1))
            set_gcoeff(A, j+1, i+1, x)  # A[j+1, i+1] = x (using 1-based indexing)
    return A


cdef GEN _new_GEN_from_fmpq_mat_t_rotate90(fmpq_mat_t B) noexcept:
    r"""
    Create a new PARI ``t_MAT`` with ``nr`` rows and ``nc`` columns
    from a ``fmpq_mat_t`` and rotate the matrix 90 degrees
    counterclockwise.  So the resulting matrix will have ``nc`` rows
    and ``nr`` columns.

    For internal use only; this directly uses the PARI stack.
    One should call ``sig_on()`` before and ``sig_off()`` after.
    """
    cdef GEN x
    cdef GEN A = zeromatcopy(fmpq_mat_ncols(B), fmpq_mat_nrows(B))
    cdef Py_ssize_t i, j
    for i in range(fmpq_mat_nrows(B)):
        for j in range(fmpq_mat_ncols(B)):
            x = _new_GEN_from_fmpq_t(fmpq_mat_entry(B, i, fmpq_mat_ncols(B) - j - 1))
            set_gcoeff(A, j+1, i+1, x)  # A[j+1, i+1] = x (using 1-based indexing)
    return A


cdef Gen integer_matrix(fmpz_mat_t B, bint rotate):
    """
    EXAMPLES::

        sage: matrix(ZZ,2,[1..6]).__pari__()   # indirect doctest
        [1, 2, 3; 4, 5, 6]
    """
    cdef GEN g
    sig_on()
    if rotate:
        g = _new_GEN_from_fmpz_mat_t_rotate90(B)
    else:
        g = _new_GEN_from_fmpz_mat_t(B)
    return new_gen(g)


cdef Gen rational_matrix(fmpq_mat_t B, bint rotate):
    """
    EXAMPLES::

        sage: matrix(QQ, 2, [1/2, 2/3, 3/4, 4/5]).__pari__() # indirect doctest
        [1/2, 2/3; 3/4, 4/5]
    """
    cdef GEN g
    sig_on()
    if rotate:
        g = _new_GEN_from_fmpq_mat_t_rotate90(B)
    else:
        g = _new_GEN_from_fmpq_mat_t(B)
    return new_gen(g)



r"""
Ariki-Koike Algebra Representations

AUTHORS:

- Travis Scrimshaw (2023-12-28): Initial version
"""

#*****************************************************************************
#       Copyright (C) 2023 Travis Scrimshaw <tcscrims at gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.misc.misc_c import prod
from sage.misc.latex import latex
from sage.categories.modules import Modules
from sage.rings.integer_ring import ZZ
from sage.combinat.free_module import CombinatorialFreeModule
from sage.combinat.partition_tuple import PartitionTuples
from sage.combinat.permutation import Permutations
from sage.combinat.tableau_tuple import StandardTableauTuples


class SpechtModule(CombinatorialFreeModule):
    r"""
    Specht module of the Ariki-Koike algebra.

    Let `H_{r,n}(q, u)` be the Ariki-Koike algebra with parameters `q`
    and `u = (u_1, \ldots, u_r)` (note our indexing convention for
    the `u` parameters differs from
    :mod:`sage.algebras.hecke_algebras.ariki_koike_algebra`) over a
    commutative ring `R`. Let `\lambda` be a partition tuple of level
    `r` and size `n`. The *Specht module* of shape `\lambda` is the (right)
    `H_{r,n}(q,u)`-representation `S^{\lambda}` given as free `R`-module
    with basis given by the standard tableau (tuples) of shape `\lambda`.

    We will now describe the right action of the Ariki-Koike algebra,
    but we first need to set some notation and definitions.
    Let `t` be a standard tableau tuple of level `r` and size `n`.
    Define the *residue* of `i` in `t` to be `r_t(i) = q^{c-r} u_k`,
    where `i` is in cell `(r, c)` of the `k`-th tableau.

    The action of `L_i` is given by `t \cdot L_i = r_T(i) t`. For `T_i`,
    we need to consider the following cases. If `i, i+1` are in the same
    row (resp. column), then `t \cdot T_i = q t` (resp. `t \cdot T_i = -t`).
    Otherwise if we swap `i, i+1`, the resulting tableau tuple `s` is again
    standard and the action is given by

    .. MATH::

        t \cdot T_i = \frac{(q - 1) r_t(i)}{r_s(i) - r_t(i)} t
        + \frac{q r_t(i) - r_s(i)}{r_s(i) - r_t(i)} s.

    Note that `r_s(i) = r_t(i+1)`.

    Over a field of characteristic `0`, the set of Specht modules for all
    partition tuples of level `r` and size `n` form the complete set
    of irreducible modules for `H_{r,n}(q, u)` [AK1994]_. (The condition
    on the base ring can be weakened; see Theorem 3.2 of [Mathas2002]_.)

    EXAMPLES:

    We construct the Specht module `S^{(2,1,21)}` for `H_{3,6}(q, u)` with
    generic parameters `q, u` over `\GF(3)` and perform some basic
    computations. We change the tableaux to use the compact representation
    to condense the output::

        sage: TableauTuples.options.display = 'compact'

        sage: R = PolynomialRing(GF(3), 'u', 3)
        sage: u = R.gens()
        sage: q = R['q'].gen()
        sage: H = algebras.ArikiKoike(3, 6, q, u, use_fraction_field=True)
        sage: LT = H.LT()
        sage: T0, T1, T2, T3, T4, T5 = LT.T()
        sage: S = H.specht_module([[2], [1], [2,1]])
        sage: S.dimension()
        120
        sage: elt = S.an_element(); elt
        S[1,2|3|4,5/6] - S[1,3|2|4,5/6] + S[1,3|4|2,5/6]
        sage: elt * LT.L(3)
        u1*S[1,2|3|4,5/6] + (-u0*q)*S[1,3|2|4,5/6] + u0*q*S[1,3|4|2,5/6]
        sage: elt * T2
        (((-u0-u1)*q-u1)/(-u0*q+u1))*S[1,2|3|4,5/6]
         + (((-u0+u2)*q)/(u0*q-u2))*S[1,2|4|3,5/6]
         + ((-u0*q^2-u0*q-u1)/(-u0*q+u1))*S[1,3|2|4,5/6]
         + ((u0*q^2-u0*q)/(u0*q-u2))*S[1,3|4|2,5/6]
        sage: (elt * T3) * T2 == elt * (T3 * T2)
        True
        sage: elt * T2 * T3 * T2 == elt * T3 * T2 * T3
        True
        sage: elt * T0 * T1 * T0 * T1 == elt * T1 * T0 * T1 * T0
        True
        sage: elt * T2 * T5 == elt * T5 * T2
        True

        sage: TableauTuples.options._reset()

    REFERENCES:

    - [AK1994]_
    - [DJM1998]_
    - [DR2001]_
    - [Mathas2002]_
    - [Mathas2004]_
    """
    @staticmethod
    def __classcall_private__(cls, AK, la):
        """
        Normalize input to ensure a unique representation.

        EXAMPLES::

            sage: AK = algebras.ArikiKoike(3, 6)
            sage: S1 = AK.specht_module([[3], [1], [1,1]])
            sage: S2 = AK.specht_module(PartitionTuple([[3], [1], [1,1]]))
            sage: S1 is S2
            True
        """
        la = PartitionTuples(AK._r, AK._n)(la)
        return super().__classcall__(cls, AK, la)

    def __init__(self, AK, la):
        r"""
        Initialize ``self``.

        EXAMPLES::

            sage: AK = algebras.ArikiKoike(3, 6, use_fraction_field=True)
            sage: S = AK.specht_module([[3], [1], [1,1]])
            sage: TestSuite(S).run()  # long time
            sage: Sp = AK.specht_module([[], [2,1,1], [2]])
            sage: TestSuite(Sp).run()  # long time
        """
        self._shape = la
        self._AK = AK
        self._q = AK.q()
        self._u = AK.u()
        self._Pn = Permutations(la.size())
        indices = StandardTableauTuples(la)
        R = AK.base_ring()
        cat = Modules(R).FiniteDimensional().WithBasis()
        CombinatorialFreeModule.__init__(self, R, indices, category=cat, prefix='S')

    def _repr_(self) -> str:
        r"""
        Return a string representation of ``self``.

        EXAMPLES::

            sage: AK = algebras.ArikiKoike(3, 8)
            sage: AK.specht_module([[3], [], [2,2,1]])
            Specht module of shape ([3], [], [2, 2, 1]) for
             Ariki-Koike algebra of rank 3 and order 8 with q=q and u=(u0, u1, u2)
              over ... over Integer Ring
        """
        return "Specht module of shape {} for {}".format(self._shape, self._AK)

    def _latex_(self) -> str:
        r"""
        Return a latex representation of ``self``.

        EXAMPLES::

            sage: AK = algebras.ArikiKoike(3, 8)
            sage: S = AK.specht_module([[3], [], [2,2,1]])
            sage: latex(S)
            S^{{\def\lr#1{\multicolumn{1}{|@{\hspace{.6ex}}c@{\hspace{.6ex}}|}{\raisebox{-.3ex}{$#1$}}}
            \raisebox{-.6ex}{$\begin{array}[b]{*{3}c}\cline{1-3}
            \lr{\phantom{x}}&\lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-3}
            \end{array}$},\emptyset,\raisebox{-.6ex}{$\begin{array}[b]{*{2}c}\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}&\lr{\phantom{x}}\\\cline{1-2}
            \lr{\phantom{x}}\\\cline{1-1}
            \end{array}$}
            }}_{\mathcal{H}_{3,8}(q)}
        """
        return "S^{{{}}}_{{{}}}".format(latex(self._shape), latex(self._AK))

    def _test_representation(self, **options):
        r"""
        Test that the relations of the Ariki-Koike algebra are satisfied.

        EXAMPLES::

            sage: q = ZZ['q'].fraction_field().gen()
            sage: AK = algebras.ArikiKoike(2, 4, q, [q^2+1, q-3], q.parent())
            sage: S = AK.specht_module([[2,1], [1]])
            sage: S._test_representation(elements=S.basis())
        """
        tester = self._tester(**options)
        n = self._shape.size()
        q = self._q
        from sage.misc.misc import some_tuples
        from sage.rings.polynomial.polynomial_ring_constructor import PolynomialRing

        # Build the polynomial for testing the T0 action
        z = PolynomialRing(self.base_ring(), 'DUMMY').gen()
        T0_poly = -prod(z - val for val in self._u)

        def apply_T0_power(b, exp):
            for i in range(exp):
                b = b.T(0)
            return b

        AKelts = self._AK.some_elements()
        for b in tester.some_elements():
            t0 = self.linear_combination((apply_T0_power(b, exp), c)
                                         for exp, c in enumerate(T0_poly))
            tester.assertEqual(t0, self.zero())

            tester.assertEqual(b.T([0, 1, 0, 1]), b.T([1, 0, 1, 0]))
            tester.assertEqual(b.T(1).T(1), (q-1)*b.T(1) + q*b)
            for i in range(2, n):
                tester.assertEqual(b.T(i).T(i), (q-1)*b.T(i) + q*b)
                tester.assertEqual(b.T(i).T(0), b.T(0).T(i))
                if i < n - 1:
                    tester.assertEqual(b.T([i, i+1, i]), b.T([i+1, i, i+1]))
                    for j in range(i+2, n):
                        tester.assertEqual(b.T([i, j]), b.T([j, i]))

            for (x, y) in some_tuples(AKelts, 2, tester._max_runs):
                tester.assertEqual(b*(x*y), (b*x)*y)

    def _L_on_basis(self, i, t):
        """
        Return the action of `L_i` on the basis element indexed by
        the standard tableau tuple ``t``.

        EXAMPLES::

            sage: AK = algebras.ArikiKoike(3, 10)
            sage: S = AK.specht_module([[2,1], [], [3,2,2]])
            sage: P = S.basis().keys()
            sage: t = P([[[2,4],[8]], [], [[1,3,7],[5,6],[9,10]]])
            sage: S._L_on_basis(1, t)
            u2*S[([[2, 4], [8]], [], [[1, 3, 7], [5, 6], [9, 10]])]
            sage: S._L_on_basis(4, t)
            u0*q*S[([[2, 4], [8]], [], [[1, 3, 7], [5, 6], [9, 10]])]
            sage: S._L_on_basis(6, t)
            u2*S[([[2, 4], [8]], [], [[1, 3, 7], [5, 6], [9, 10]])]
            sage: S._L_on_basis(8, t)
            (u0*q^-1)*S[([[2, 4], [8]], [], [[1, 3, 7], [5, 6], [9, 10]])]
            sage: S._L_on_basis(9, t)
            (u2*q^-2)*S[([[2, 4], [8]], [], [[1, 3, 7], [5, 6], [9, 10]])]
        """
        c = t.cells_containing(i)[0]
        if len(c) == 2:  # it is of level 1 and a regular tableau
            c = (0,) + c
        res = self._q**(c[2]-c[1]) * self._u[c[0]]
        R = self.base_ring()
        return self.element_class(self, {t: R(res)})

    def _T_on_basis(self, i, t):
        r"""
        Return the action of `T_i` on the basis element indexed by
        the standard tableau tuple ``t``.

        EXAMPLES::

            sage: AK = algebras.ArikiKoike(3, 10, use_fraction_field=True)
            sage: S = AK.specht_module([[2,1], [], [3,2,2]])
            sage: P = S.basis().keys()
            sage: t = P([[[2,4],[8]], [], [[1,5,7],[3,6],[9,10]]])
            sage: S._T_on_basis(0, t) == S._L_on_basis(1, t)
            True
            sage: S._T_on_basis(1, t)
            ((u2*q-u0)/(u0-u2))*S[([[1, 4], [8]], [], [[2, 5, 7], [3, 6], [9, 10]])]
             + ((u0*q-u0)/(u0-u2))*S[([[2, 4], [8]], [], [[1, 5, 7], [3, 6], [9, 10]])]
            sage: S._T_on_basis(2, t)
            ((u2*q-u2)/(-u0*q+u2))*S[([[2, 4], [8]], [], [[1, 5, 7], [3, 6], [9, 10]])]
             + ((u0*q^2-u2)/(-u0*q+u2))*S[([[3, 4], [8]], [], [[1, 5, 7], [2, 6], [9, 10]])]
            sage: S._T_on_basis(5, t)
            -S[([[2, 4], [8]], [], [[1, 5, 7], [3, 6], [9, 10]])]
            sage: S._T_on_basis(7, t)
            ((u2*q^4-u0)/(-u2*q^3+u0))*S[([[2, 4], [7]], [], [[1, 5, 8], [3, 6], [9, 10]])]
             + ((u0*q-u0)/(-u2*q^3+u0))*S[([[2, 4], [8]], [], [[1, 5, 7], [3, 6], [9, 10]])]
            sage: S._T_on_basis(9, t)
            q*S[([[2, 4], [8]], [], [[1, 5, 7], [3, 6], [9, 10]])]
        """
        R = self.base_ring()
        if i == 0:
            return self._L_on_basis(1, t)

        ct = t.cells_containing(i)[0]
        cs = t.cells_containing(i+1)[0]
        if len(ct) == 2:  # it is of level 1 and a regular tableau
            ct = (0,) + ct
            cs = (0,) + cs

        if ct[0] == cs[0] and ct[2] == cs[2]:  # same column
            return self.element_class(self, {t: -R.one()})

        if ct[0] == cs[0] and ct[1] == cs[1]:  # same row
            return self.element_class(self, {t: self._q})

        # result is standard
        s = t.symmetric_group_action_on_entries(self._Pn.simple_reflection(i))
        assert s.parent() is t.parent()

        def res(cell):
            return self._q**(cell[2]-cell[1]) * self._u[cell[0]]

        # Note that the residue of i in t is given by the cell c
        #   and of i in s corresponds to cell cp because the
        #   corresponding action of the permutation on t.
        one = self.base_ring().one()
        denom = res(cs) - res(ct)
        coefft = (self._q - one) * res(cs) / denom
        coeffs = (self._q * res(ct) - res(cs)) / denom
        return self.element_class(self, {t: R(coefft), s: R(coeffs)})

    def ariki_koike_algebra(self):
        r"""
        Return the Ariki-Koike algebra that ``self`` is a representation of.

        EXAMPLES::

            sage: AK = algebras.ArikiKoike(3, 6)
            sage: S = AK.specht_module([[2], [], [3,1]])
            sage: S.ariki_koike_algebra() is AK
            True
        """
        return self._AK

    class Element(CombinatorialFreeModule.Element):
        def _acted_upon_(self, scalar, self_on_left):
            r"""
            Return the action of ``scalar`` on ``self``.

            EXAMPLES::

                sage: TableauTuples.options.display = 'compact'
                sage: AK = algebras.ArikiKoike(4, 6, use_fraction_field=True)
                sage: q = AK.q()
                sage: LT = AK.LT()
                sage: T = AK.T()
                sage: S = AK.specht_module([[], [2], [1], [2,1]])
                sage: elt = S.an_element()
                sage: 5 * elt
                5*S[-|1,2|3|4,5/6] + 10*S[-|1,3|2|4,5/6] + 5*S[-|1,3|4|2,5/6]
                 + 15*S[-|2,3|1|4,5/6]
                sage: elt * (q - 2)
                (q-2)*S[-|1,2|3|4,5/6] + (2*q-4)*S[-|1,3|2|4,5/6]
                 + (q-2)*S[-|1,3|4|2,5/6] + (3*q-6)*S[-|2,3|1|4,5/6]
                sage: elt * LT.an_element() == elt * T(LT.an_element())
                True
                sage: T.an_element() * elt
                Traceback (most recent call last):
                ...
                TypeError: unsupported operand parent(s) for *: 'Ariki-Koike algebra ... over Integer Ring'
                sage: TableauTuples.options._reset()

            TESTS::

                sage: AK = algebras.ArikiKoike(2, 4, use_fraction_field=True)
                sage: LT = AK.LT()
                sage: T = AK.T()
                sage: S = AK.specht_module([[1], [2,1]])
                sage: B = list(LT.basis())[::55]
                sage: all(b * x == b * T(x) for b in S.basis() for x in B)  # long time
                True
            """
            ret = super()._acted_upon_(scalar, self_on_left)
            if ret is not None:
                return ret
            if not self_on_left:  # only a right action
                return None
            P = self.parent()
            if scalar not in P._AK:
                return None
            scalar = P._AK(scalar)
            if scalar.parent() is P._AK.LT():
                return P.linear_combination((self.L(sum(([i]*val for i, val in enumerate(m[0], start=1)), [])).T(m[1].reduced_word()), c)
                                            for m, c in scalar)
            if scalar.parent() is P._AK.T():
                AKT = P._AK.T()
                return P.linear_combination((self.T(AKT._basis_to_word(m)), c)
                                            for m, c in scalar)
            return self * P._AK.LT()(scalar)

        def L(self, i):
            r"""
            Return the (right) action of `L_i` on ``self``.

            INPUT:

            - ``i`` -- an integer or a list of integers

            EXAMPLES::

                sage: TableauTuples.options.display = 'compact'  # compact tableau printing
                sage: AK = algebras.ArikiKoike(3, 6, use_fraction_field=True)
                sage: S = AK.specht_module([[2], [], [3,1]])
                sage: elt = S.an_element(); elt
                S[1,2|-|3,4,5/6] + 2*S[1,3|-|2,4,5/6]
                 + 3*S[2,3|-|1,4,5/6] + S[2,4|-|1,3,5/6]
                sage: elt.L(1)
                u0*S[1,2|-|3,4,5/6] + 2*u0*S[1,3|-|2,4,5/6]
                 + 3*u2*S[2,3|-|1,4,5/6] + u2*S[2,4|-|1,3,5/6]
                sage: elt.L(2)
                u0*q*S[1,2|-|3,4,5/6] + 2*u2*S[1,3|-|2,4,5/6]
                 + 3*u0*S[2,3|-|1,4,5/6] + u0*S[2,4|-|1,3,5/6]
                sage: elt.L(6)
                u2/q*S[1,2|-|3,4,5/6] + 2*u2/q*S[1,3|-|2,4,5/6]
                 + 3*u2/q*S[2,3|-|1,4,5/6] + u2/q*S[2,4|-|1,3,5/6]
                sage: elt.L([3,3,3])
                u2^3*S[1,2|-|3,4,5/6] + 2*u0^3*q^3*S[1,3|-|2,4,5/6]
                + 3*u0^3*q^3*S[2,3|-|1,4,5/6] + u2^3*q^3*S[2,4|-|1,3,5/6]
                sage: LT = AK.LT()
                sage: elt.L([3,3,3]) == elt * (LT.L(3)^3)
                True
                sage: TableauTuples.options._reset()  # reset
            """
            if not self:  # action on 0 is 0
                return self
            if i not in ZZ:
                ret = self
                for val in i:
                    ret = ret.L(val)
                return ret
            P = self.parent()
            return P.linear_combination((P._L_on_basis(i, t), c) for t, c in self)

        def T(self, i):
            r"""
            Return the (right) action of `T_i` on ``self``.

            INPUT:

            - ``i`` -- an integer or a list of integers

            EXAMPLES::

                sage: TableauTuples.options.display = 'compact'  # compact tableau printing
                sage: AK = algebras.ArikiKoike(3, 10, use_fraction_field=True)
                sage: q = AK.q()
                sage: S = AK.specht_module([[2,1], [], [3,2,2]])
                sage: P = S.basis().keys()
                sage: t = P([[[2,4],[8]], [], [[1,5,7],[3,6],[9,10]]])
                sage: b = S.basis()[t]
                sage: b.T(2)
                ((u2*q-u2)/(-u0*q+u2))*S[2,4/8|-|1,5,7/3,6/9,10]
                 + ((u0*q^2-u2)/(-u0*q+u2))*S[3,4/8|-|1,5,7/2,6/9,10]
                sage: b.T(6)
                ((-q)/(q+1))*S[2,4/8|-|1,5,6/3,7/9,10]
                 + (q^2/(q+1))*S[2,4/8|-|1,5,7/3,6/9,10]
                sage: b.T([2,1,2]) == b.T([1,2,1])
                True
                sage: b.T(9)
                q*S[2,4/8|-|1,5,7/3,6/9,10]
                sage: all(b.T([i,i]) == (q-1)*b.T(i) + q*b for i in range(1,10))
                True
                sage: b.T(0)
                u2*S[2,4/8|-|1,5,7/3,6/9,10]
                sage: b.T([0,1,0,1]) == b.T([1,0,1,0])
                True
                sage: TableauTuples.options._reset()  # reset
            """
            if not self:  # action on 0 is 0
                return self
            if i not in ZZ:
                ret = self
                for val in i:
                    ret = ret.T(val)
                return ret
            P = self.parent()
            return P.linear_combination((P._T_on_basis(i, t), c) for t, c in self)
